@@ -115,6 +115,9 @@ class EncodedString(_unicode):
     # otherwise
     encoding = None
 
+    def __deepcopy__(self, memo):
+        return self
+
     def byteencode(self):
         assert self.encoding is not None
         return self.encode(self.encoding)
@@ -131,6 +134,9 @@ class BytesLiteral(_bytes):
     # bytes subclass that is compatible with EncodedString
     encoding = None
 
+    def __deepcopy__(self, memo):
+        return self
+
     def byteencode(self):
         if IS_PYTHON3:
             return _bytes(self)
@@ -232,9 +238,23 @@ def escape_byte_string(s):
                 append(c)
         return join_bytes(l).decode('ISO-8859-1')
 
-def split_string_literal(s):
+def split_string_literal(s, limit=2000):
     # MSVC can't handle long string literals.
-    if len(s) < 2047:
+    if len(s) < limit:
         return s
     else:
-        return '""'.join([s[i:i+2000] for i in range(0, len(s), 2000)]).replace(r'\""', '""\\')
+        start = 0
+        chunks = []
+        while start < len(s):
+            end = start + limit
+            if len(s) > end-4 and '\\' in s[end-4:end]:
+                end -= 4 - s[end-4:end].find('\\') # just before the backslash
+                while s[end-1] == '\\':
+                    end -= 1
+                    if end == start:
+                        # must have been a long line of backslashes
+                        end = start + limit - (limit % 2) - 4
+                        break
+            chunks.append(s[start:end])
+            start = end
+        return '""'.join(chunks)
