@@ -3,23 +3,16 @@
 #
 
 import re
-from Cython import Utils
 from Errors import warning, error, InternalError
 from StringEncoding import EncodedString
 import Options, Naming
 import PyrexTypes
 from PyrexTypes import py_object_type, unspecified_type
-import TypeSlots
 from TypeSlots import \
     pyfunction_signature, pymethod_signature, \
     get_special_method_signature, get_property_accessor_signature
 import Code
 import __builtin__ as builtins
-try:
-    set
-except NameError:
-    from sets import Set as set
-import copy
 
 possible_identifier = re.compile(ur"(?![0-9])\w+$", re.U).match
 nice_identifier = re.compile('^[a-zA-Z0-0_]+$').match
@@ -44,12 +37,9 @@ def c_safe_identifier(cname):
 class BufferAux(object):
     writable_needed = False
 
-    def __init__(self, buffer_info_var, stridevars, shapevars,
-                 suboffsetvars):
-        self.buffer_info_var = buffer_info_var
-        self.stridevars = stridevars
-        self.shapevars = shapevars
-        self.suboffsetvars = suboffsetvars
+    def __init__(self, buflocal_nd_var, rcbuf_var):
+        self.buflocal_nd_var = buflocal_nd_var
+        self.rcbuf_var = rcbuf_var
 
     def __repr__(self):
         return "<BufferAux %r>" % self.__dict__
@@ -73,7 +63,10 @@ class Entry(object):
     # is_variable      boolean    Is a variable
     # is_cfunction     boolean    Is a C function
     # is_cmethod       boolean    Is a C method of an extension type
+    # is_builtin_cmethod boolean  Is a C method of a builtin type (implies is_cmethod)
     # is_unbound_cmethod boolean  Is an unbound C method of an extension type
+    # is_final_cmethod   boolean  Is non-overridable C method
+    # is_inline_cmethod  boolean  Is inlined C method
     # is_anonymous     boolean    Is a anonymous pyfunction entry
     # is_type          boolean    Is a type definition
     # is_cclass        boolean    Is an extension class
@@ -119,7 +112,6 @@ class Entry(object):
     # buffer_aux       BufferAux or None  Extra information needed for buffer variables
     # inline_func_in_pxd boolean  Hacky special case for inline function in pxd file.
     #                             Ideally this should not be necesarry.
-    # assignments      [ExprNode] List of expressions that get assigned to this entry.
     # might_overflow   boolean    In an arithmetic expression that could cause
     #                             overflow (used for type inference).
     # utility_code_definition     For some Cython builtins, the utility code
@@ -127,6 +119,11 @@ class Entry(object):
     #                             Currently only supported for CythonScope entries.
     # error_on_uninitialized      Have Control Flow issue an error when this entry is
     #                             used uninitialized
+    # cf_used          boolean    Entry is used
+    # is_fused_specialized boolean Whether this entry of a cdef or def function
+    #                              is a specialization
+
+    # TODO: utility_code and utility_code_definition serves the same purpose...
 
     inline_func_in_pxd = False
     borrowed = 0
@@ -140,7 +137,10 @@ class Entry(object):
     is_variable = 0
     is_cfunction = 0
     is_cmethod = 0
+    is_builtin_cmethod = False
     is_unbound_cmethod = 0
+    is_final_cmethod = 0
+    is_inline_cmethod = 0
     is_anonymous = 0
     is_type = 0
     is_cclass = 0
@@ -157,8 +157,10 @@ class Entry(object):
     from_closure = 0
     is_declared_generic = 0
     is_readonly = 0
+    pyfunc_cname = None
     func_cname = None
     func_modifiers = []
+    final_func_cname = None
     doc = None
     as_variable = None
     xdecref_cleanup = 0
@@ -178,10 +180,14 @@ class Entry(object):
     buffer_aux = None
     prev_entry = None
     might_overflow = 0
+    fused_cfunction = None
+    is_fused_specialized = False
     utility_code_definition = None
+    needs_property = False
     in_with_gil_block = 0
     from_cython_utility_code = None
     error_on_uninitialized = False
+    cf_used = True
 
     def __init__(self, name, cname, type, pos = None, init = None):
         self.name = name
@@ -190,7 +196,6 @@ class Entry(object):
         self.pos = pos
         self.init = init
         self.overloaded_alternatives = []
-        self.assignments = []
         self.cf_assignments = []
         self.cf_references = []
 
@@ -247,6 +252,7 @@ class Scope(object):
     scope_prefix = ""
     in_cinclude = 0
     nogil = 0
+    fused_to_specific = None
 
     def __init__(self, name, outer_scope, parent_scope):
         # The outer_scope is the next scope in the lookup chain.
@@ -283,27 +289,31 @@ class Scope(object):
         self.return_type = None
         self.id_counters = {}
 
-    def merge_in(self, other, merge_unused=True):
+    def __deepcopy__(self, memo):
+        return self
+
+    def merge_in(self, other, merge_unused=True, whitelist=None):
         # Use with care...
-        entries = [(name, entry)
-                       for name, entry in other.entries.iteritems()
-                           if entry.used or merge_unused]
-        # !@#$ py23
-        entries = dict(entries)
+        entries = []
+        for name, entry in other.entries.iteritems():
+            if not whitelist or name in whitelist:
+                if entry.used or merge_unused:
+                    entries.append((name, entry))
 
         self.entries.update(entries)
 
         for attr in ('const_entries',
-                  'type_entries',
-                  'sue_entries',
-                  'arg_entries',
-                  'var_entries',
-                  'pyfunc_entries',
-                  'cfunc_entries',
-                  'c_class_entries'):
+                     'type_entries',
+                     'sue_entries',
+                     'arg_entries',
+                     'var_entries',
+                     'pyfunc_entries',
+                     'cfunc_entries',
+                     'c_class_entries'):
             self_entries = getattr(self, attr)
+            names = set([e.name for e in self_entries])
             for entry in getattr(other, attr):
-                if entry.used or merge_unused:
+                if (entry.used or merge_unused) and entry.name not in names:
                     self_entries.append(entry)
 
     def __str__(self):
@@ -325,6 +335,11 @@ class Scope(object):
         return self.mangle(prefix)
         #return self.parent_scope.mangle(prefix, self.name)
 
+    def mangle_class_private_name(self, name):
+        if self.parent_scope:
+            return self.parent_scope.mangle_class_private_name(name)
+        return name
+
     def next_id(self, name=None):
         # Return a cname fragment that is unique for this module
         counters = self.global_scope().id_counters
@@ -349,12 +364,12 @@ class Scope(object):
         # Return the module-level scope containing this scope.
         return self.outer_scope.builtin_scope()
 
-    def declare(self, name, cname, type, pos, visibility, shadow = 0):
+    def declare(self, name, cname, type, pos, visibility, shadow = 0, is_type = 0):
         # Create new entry, and add to dictionary if
         # name is not None. Reports a warning if already
         # declared.
-        if type.is_buffer and not isinstance(self, LocalScope):
-            error(pos, ERR_BUF_LOCALONLY)
+        if type.is_buffer and not isinstance(self, LocalScope): # and not is_type:
+            error(pos, 'Buffer types only allowed as function local variables')
         if not self.in_cinclude and cname and re.match("^_[_A-Z]+$", cname):
             # See http://www.gnu.org/software/libc/manual/html_node/Reserved-Names.html#Reserved-Names
             warning(pos, "'%s' is a reserved name in C." % cname, -1)
@@ -374,6 +389,11 @@ class Scope(object):
 #                entries[name] = entry
             if not shadow:
                 entries[name] = entry
+
+        if type.is_memoryviewslice:
+            import MemoryView
+            entry.init = MemoryView.memslice_entry_init
+
         entry.scope = self
         entry.visibility = visibility
         return entry
@@ -394,15 +414,21 @@ class Scope(object):
         return entry
 
     def declare_type(self, name, type, pos,
-            cname = None, visibility = 'private', api = 0, defining = 1, shadow = 0):
+            cname = None, visibility = 'private', api = 0, defining = 1,
+            shadow = 0, template = 0):
         # Add an entry for a type definition.
         if not cname:
             cname = name
-        entry = self.declare(name, cname, type, pos, visibility, shadow)
+        entry = self.declare(name, cname, type, pos, visibility, shadow,
+                             is_type=True)
         entry.is_type = 1
         entry.api = api
         if defining:
             self.type_entries.append(entry)
+
+        if not template:
+            type.entry = entry
+
         # here we would set as_variable to an object representing this type
         return entry
 
@@ -458,18 +484,21 @@ class Scope(object):
         return entry
 
     def declare_cpp_class(self, name, scope,
-            pos, cname = None, base_classes = [],
+            pos, cname = None, base_classes = (),
             visibility = 'extern', templates = None):
-        if visibility != 'extern':
-            error(pos, "C++ classes may only be extern")
         if cname is None:
-            cname = name
+            if self.in_cinclude or (visibility != 'private'):
+                cname = name
+            else:
+                cname = self.mangle(Naming.type_prefix, name)
+        base_classes = list(base_classes)
         entry = self.lookup_here(name)
         if not entry:
             type = PyrexTypes.CppClassType(
                 name, scope, cname, base_classes, templates = templates)
             entry = self.declare_type(name, type, pos, cname,
                 visibility = visibility, defining = scope is not None)
+            self.sue_entries.append(entry)
         else:
             if not (entry.is_type and entry.type.is_cpp_class):
                 error(pos, "'%s' redeclared " % name)
@@ -481,17 +510,13 @@ class Scope(object):
                     entry.type.scope = scope
                     self.type_entries.append(entry)
             if base_classes:
-                if entry.type.base_classes and not entry.type.base_classes == base_classes:
+                if entry.type.base_classes and entry.type.base_classes != base_classes:
                     error(pos, "Base type does not match previous declaration")
                 else:
                     entry.type.base_classes = base_classes
             if templates or entry.type.templates:
                 if templates != entry.type.templates:
                     error(pos, "Template parameters do not match previous declaration")
-        if templates is not None and entry.type.scope is not None:
-            for T in templates:
-                template_entry = entry.type.scope.declare(T.name, T.name, T, None, 'extern')
-                template_entry.is_type = 1
 
         def declare_inherited_attributes(entry, base_classes):
             for base_class in base_classes:
@@ -502,6 +527,7 @@ class Scope(object):
                     entry.type.scope.declare_inherited_cpp_attributes(base_class.scope)
         if entry.type.scope:
             declare_inherited_attributes(entry, base_classes)
+            entry.type.scope.declare_var(name="this", cname="this", type=PyrexTypes.CPtrType(entry.type), pos=entry.pos)
         if self.is_cpp_class_scope:
             entry.type.namespace = self.outer_scope.lookup(self.name).type
         return entry
@@ -545,7 +571,7 @@ class Scope(object):
         if type.is_cpp_class and visibility != 'extern':
             constructor = type.scope.lookup(u'<init>')
             if constructor is not None and PyrexTypes.best_match([], constructor.all_alternatives()) is None:
-                error(pos, "C++ class must have a default constructor to be stack allocated")
+                error(pos, "C++ class must have a no-arg constructor to be stack allocated")
         entry = self.declare(name, cname, type, pos, visibility)
         entry.is_variable = 1
         if in_pxd and visibility != 'extern':
@@ -657,7 +683,10 @@ class Scope(object):
             entry.is_implemented = True
         if modifiers:
             entry.func_modifiers = modifiers
-        entry.utility_code = utility_code
+        if utility_code:
+            assert not entry.utility_code, "duplicate utility code definition in entry %s (%s)" % (name, cname)
+            entry.utility_code = utility_code
+        type.entry = entry
         return entry
 
     def add_cfunction(self, name, type, pos, cname, visibility, modifiers):
@@ -714,6 +743,8 @@ class Scope(object):
     def lookup_type(self, name):
         entry = self.lookup(name)
         if entry and entry.is_type:
+            if entry.type.is_fused and self.fused_to_specific:
+                return entry.type.specialize(self.fused_to_specific)
             return entry.type
 
     def lookup_operator(self, operator, operands):
@@ -729,6 +760,13 @@ class Scope(object):
             return None
         return PyrexTypes.best_match(operands, function.all_alternatives())
 
+    def lookup_operator_for_types(self, pos, operator, types):
+        from Nodes import Node
+        class FakeOperand(Node):
+            pass
+        operands = [FakeOperand(pos, type=type) for type in types]
+        return self.lookup_operator(operator, operands)
+
     def use_utility_code(self, new_code):
         self.global_scope().use_utility_code(new_code)
 
@@ -758,6 +796,24 @@ class Scope(object):
     def add_include_file(self, filename):
         self.outer_scope.add_include_file(filename)
 
+    def get_refcounted_entries(self, include_weakref=False):
+        py_attrs = []
+        py_buffers = []
+        memoryview_slices = []
+
+        for entry in self.var_entries:
+            if entry.type.is_pyobject:
+                if include_weakref or entry.name != "weakref":
+                    py_attrs.append(entry)
+            elif entry.type == PyrexTypes.c_py_buffer_type:
+                py_buffers.append(entry)
+            elif entry.type.is_memoryviewslice:
+                memoryview_slices.append(entry)
+
+        have_entries = py_attrs or py_buffers or memoryview_slices
+        return have_entries, (py_attrs, py_buffers, memoryview_slices)
+
+
 class PreImportScope(Scope):
 
     namespace_cname = Naming.preimport_cname
@@ -830,7 +886,7 @@ class BuiltinScope(Scope):
         scope = CClassScope(name, outer_scope=None, visibility='extern')
         scope.directives = {}
         if name == 'bool':
-            scope.directives['final'] = True
+            type.is_final_type = True
         type.set_scope(scope)
         self.type_names[name] = 1
         entry = self.declare_type(name, type, None, visibility='extern')
@@ -907,6 +963,7 @@ class ModuleScope(Scope):
     # has_import_star      boolean            Module contains import *
     # cpp                  boolean            Compiling a C++ file
     # is_cython_builtin    boolean            Is this the Cython builtin scope (or a child scope)
+    # is_package           boolean            Is this a package module? (__init__)
 
     is_module_scope = 1
     has_import_star = 0
@@ -917,12 +974,14 @@ class ModuleScope(Scope):
         self.parent_module = parent_module
         outer_scope = Builtin.builtin_scope
         Scope.__init__(self, name, outer_scope, parent_module)
-        if name != "__init__":
-            self.module_name = name
-        else:
+        if name == "__init__":
             # Treat Spam/__init__.pyx specially, so that when Python loads
             # Spam/__init__.so, initSpam() is defined.
             self.module_name = parent_module.module_name
+            self.is_package = True
+        else:
+            self.module_name = name
+            self.is_package = False
         self.module_name = EncodedString(self.module_name)
         self.context = context
         self.module_cname = Naming.module_cname
@@ -943,8 +1002,8 @@ class ModuleScope(Scope):
         self.cached_builtins = []
         self.undeclared_cached_builtins = []
         self.namespace_cname = self.module_cname
-        for name in ['__builtins__', '__name__', '__file__', '__doc__']:
-            self.declare_var(EncodedString(name), py_object_type, None)
+        for var_name in ['__builtins__', '__name__', '__file__', '__doc__', '__path__']:
+            self.declare_var(EncodedString(var_name), py_object_type, None)
 
     def qualifying_scope(self):
         return self.parent_module
@@ -1154,7 +1213,7 @@ class ModuleScope(Scope):
 
     def declare_c_class(self, name, pos, defining = 0, implementing = 0,
         module_name = None, base_type = None, objstruct_cname = None,
-        typeobj_cname = None, visibility = 'private', typedef_flag = 0, api = 0,
+        typeobj_cname = None, typeptr_cname = None, visibility = 'private', typedef_flag = 0, api = 0,
         buffer_defaults = None, shadow = 0):
         # If this is a non-extern typedef class, expose the typedef, but use
         # the non-typedef struct internally to avoid needing forward
@@ -1196,7 +1255,10 @@ class ModuleScope(Scope):
                 type.module_name = module_name
             else:
                 type.module_name = self.qualified_name
-            type.typeptr_cname = self.mangle(Naming.typeptr_prefix, name)
+            if typeptr_cname:
+                type.typeptr_cname = typeptr_cname
+            else:
+                type.typeptr_cname = self.mangle(Naming.typeptr_prefix, name)
             entry = self.declare_type(name, type, pos, visibility = visibility,
                 defining = 0, shadow = shadow)
             entry.is_cclass = True
@@ -1246,6 +1308,9 @@ class ModuleScope(Scope):
                     error(pos, "Type object name differs from previous declaration")
             type.typeobj_cname = typeobj_cname
 
+        if self.directives.get('final'):
+            entry.type.is_final_type = True
+
         # cdef classes are always exported, but we need to set it to
         # distinguish between unused Cython utility code extension classes
         entry.used = True
@@ -1513,6 +1578,10 @@ class GeneratorExpressionScope(Scope):
         self.entries[name] = entry
         return entry
 
+    def declare_pyfunction(self, name, pos, allow_redefine=False):
+        return self.outer_scope.declare_pyfunction(
+            name, pos, allow_redefine)
+
     def declare_lambda_function(self, func_cname, pos):
         return self.outer_scope.declare_lambda_function(func_cname, pos)
 
@@ -1524,8 +1593,8 @@ class ClosureScope(LocalScope):
 
     is_closure_scope = True
 
-    def __init__(self, name, scope_name, outer_scope):
-        LocalScope.__init__(self, name, outer_scope)
+    def __init__(self, name, scope_name, outer_scope, parent_scope=None):
+        LocalScope.__init__(self, name, outer_scope, parent_scope)
         self.closure_cname = "%s%s" % (Naming.closure_scope_prefix, scope_name)
 
 #    def mangle_closure_cnames(self, scope_var):
@@ -1540,6 +1609,7 @@ class ClosureScope(LocalScope):
     def declare_pyfunction(self, name, pos, allow_redefine=False):
         return LocalScope.declare_pyfunction(self, name, pos, allow_redefine, visibility='private')
 
+
 class StructOrUnionScope(Scope):
     #  Namespace of a C struct or union.
 
@@ -1578,7 +1648,7 @@ class ClassScope(Scope):
     #  Abstract base class for namespace of
     #  Python class or extension type.
     #
-    #  class_name     string   Pyrex name of the class
+    #  class_name     string   Python name of the class
     #  scope_prefix   string   Additional prefix for names
     #                          declared in the class
     #  doc    string or None   Doc string
@@ -1594,16 +1664,16 @@ class ClassScope(Scope):
             return entry
         if name == "classmethod":
             # We don't want to use the builtin classmethod here 'cause it won't do the
-            # right thing in this scope (as the class memebers aren't still functions).
+            # right thing in this scope (as the class members aren't still functions).
             # Don't want to add a cfunction to this scope 'cause that would mess with
             # the type definition, so we just return the right entry.
-            self.use_utility_code(classmethod_utility_code)
             entry = Entry(
                 "classmethod",
                 "__Pyx_Method_ClassMethod",
                 PyrexTypes.CFuncType(
                     py_object_type,
                     [PyrexTypes.CFuncTypeArg("", py_object_type, None)], 0, 0))
+            entry.utility_code_definition = Code.UtilityCode.load_cached("ClassMethod", "CythonFunction.c")
             entry.is_cfunction = 1
         return entry
 
@@ -1615,9 +1685,22 @@ class PyClassScope(ClassScope):
 
     is_py_class_scope = 1
 
+    def mangle_class_private_name(self, name):
+        return self.mangle_special_name(name)
+
+    def mangle_special_name(self, name):
+        if name and name.startswith('__') and not name.endswith('__'):
+            name = EncodedString('_%s%s' % (self.class_name.lstrip('_'), name))
+        return name
+
+    def lookup_here(self, name):
+        name = self.mangle_special_name(name)
+        return ClassScope.lookup_here(self, name)
+
     def declare_var(self, name, type, pos,
                     cname = None, visibility = 'private',
                     api = 0, in_pxd = 0, is_cdef = 0):
+        name = self.mangle_special_name(name)
         if type is unspecified_type:
             type = py_object_type
         # Add an entry for a class attribute.
@@ -1699,7 +1782,11 @@ class CClassScope(ClassScope):
                 if visibility == 'private':
                     cname = c_safe_identifier(cname)
             if type.is_cpp_class and visibility != 'extern':
-                error(pos, "C++ classes not allowed as members of an extension type, use a pointer or reference instead")
+                constructor = type.scope.lookup(u'<init>')
+                if constructor is not None and \
+                        PyrexTypes.best_match([], constructor.all_alternatives()) is None:
+                    error(pos, "C++ class must have a no-arg constructor to be a member of an extension type; use a pointer instead")
+                self.use_utility_code(Code.UtilityCode("#include <new>"))
             entry = self.declare(name, cname, type, pos, visibility)
             entry.is_variable = 1
             self.var_entries.append(entry)
@@ -1764,7 +1851,17 @@ class CClassScope(ClassScope):
     def lookup_here(self, name):
         if name == "__new__":
             name = EncodedString("__cinit__")
-        return ClassScope.lookup_here(self, name)
+        entry = ClassScope.lookup_here(self, name)
+        if entry and entry.is_builtin_cmethod:
+            if not self.parent_type.is_builtin_type:
+                # For subtypes of builtin types, we can only return
+                # optimised C methods if the type if final.
+                # Otherwise, subtypes may choose to override the
+                # method, but the optimisation would prevent the
+                # subtype method from being called.
+                if not self.parent_type.is_final_type:
+                    return None
+        return entry
 
     def declare_cfunction(self, name, type, pos,
                           cname = None, visibility = 'private', api = 0, in_pxd = 0,
@@ -1778,6 +1875,8 @@ class CClassScope(ClassScope):
             error(pos, "Self argument (%s) of C method '%s' does not match parent type (%s)" %
                   (args[0].type, name, self.parent_type))
         entry = self.lookup_here(name)
+        if cname is None:
+            cname = c_safe_identifier(name)
         if entry:
             if not entry.is_cfunction:
                 warning(pos, "'%s' redeclared  " % name, 0)
@@ -1785,10 +1884,12 @@ class CClassScope(ClassScope):
                 if defining and entry.func_cname:
                     error(pos, "'%s' already defined" % name)
                 #print "CClassScope.declare_cfunction: checking signature" ###
-                if type.same_c_signature_as(entry.type, as_cmethod = 1) and type.nogil == entry.type.nogil:
+                if entry.is_final_cmethod and entry.is_inherited:
+                    error(pos, "Overriding final methods is not allowed")
+                elif type.same_c_signature_as(entry.type, as_cmethod = 1) and type.nogil == entry.type.nogil:
                     pass
                 elif type.compatible_signature_with(entry.type, as_cmethod = 1) and type.nogil == entry.type.nogil:
-                    entry = self.add_cfunction(name, type, pos, cname or name, visibility='ignore', modifiers=modifiers)
+                    entry = self.add_cfunction(name, type, pos, cname, visibility='ignore', modifiers=modifiers)
                     defining = 1
                 else:
                     error(pos, "Signature not compatible with previous declaration")
@@ -1798,11 +1899,21 @@ class CClassScope(ClassScope):
                 error(pos,
                     "C method '%s' not previously declared in definition part of"
                     " extension type" % name)
-            entry = self.add_cfunction(name, type, pos, cname or name,
+            entry = self.add_cfunction(name, type, pos, cname,
                                        visibility, modifiers)
         if defining:
             entry.func_cname = self.mangle(Naming.func_prefix, name)
         entry.utility_code = utility_code
+        type.entry = entry
+
+        if u'inline' in modifiers:
+            entry.is_inline_cmethod = True
+
+        if (self.parent_type.is_final_type or entry.is_inline_cmethod or
+            self.directives.get('final')):
+            entry.is_final_cmethod = True
+            entry.final_func_cname = entry.func_cname
+
         return entry
 
     def add_cfunction(self, name, type, pos, cname, visibility, modifiers):
@@ -1845,18 +1956,40 @@ class CClassScope(ClassScope):
         # to work with this type.
         def adapt(cname):
             return "%s.%s" % (Naming.obj_base_cname, base_entry.cname)
-        for base_entry in \
-            base_scope.inherited_var_entries + base_scope.var_entries:
+
+        entries = base_scope.inherited_var_entries + base_scope.var_entries
+        for base_entry in entries:
                 entry = self.declare(base_entry.name, adapt(base_entry.cname),
                     base_entry.type, None, 'private')
                 entry.is_variable = 1
                 self.inherited_var_entries.append(entry)
+
+        # If the class defined in a pxd, specific entries have not been added.
+        # Ensure now that the parent (base) scope has specific entries
+        # Iterate over a copy as get_all_specialized_function_types() will mutate
+        for base_entry in base_scope.cfunc_entries[:]:
+            if base_entry.type.is_fused:
+                base_entry.type.get_all_specialized_function_types()
+
         for base_entry in base_scope.cfunc_entries:
+            cname = base_entry.cname
+            var_entry = base_entry.as_variable
+            is_builtin = var_entry and var_entry.is_builtin
+            if not is_builtin:
+                cname = adapt(cname)
             entry = self.add_cfunction(base_entry.name, base_entry.type,
-                                       base_entry.pos, adapt(base_entry.cname),
+                                       base_entry.pos, cname,
                                        base_entry.visibility, base_entry.func_modifiers)
             entry.is_inherited = 1
-
+            if base_entry.is_final_cmethod:
+                entry.is_final_cmethod = True
+                entry.is_inline_cmethod = base_entry.is_inline_cmethod
+                if (self.parent_scope == base_scope.parent_scope or
+                    entry.is_inline_cmethod):
+                    entry.final_func_cname = base_entry.final_func_cname
+            if is_builtin:
+                entry.is_builtin_cmethod = True
+                entry.as_variable = var_entry
 
 class CppClassScope(Scope):
     #  Namespace of a C++ class.
@@ -1864,24 +1997,36 @@ class CppClassScope(Scope):
     is_cpp_class_scope = 1
 
     default_constructor = None
+    type = None
 
-    def __init__(self, name, outer_scope):
+    def __init__(self, name, outer_scope, templates=None):
         Scope.__init__(self, name, outer_scope, None)
         self.directives = outer_scope.directives
         self.inherited_var_entries = []
+        if templates is not None:
+            for T in templates:
+                template_entry = self.declare(
+                    T, T, PyrexTypes.TemplatePlaceholderType(T), None, 'extern')
+                template_entry.is_type = 1
 
     def declare_var(self, name, type, pos,
                     cname = None, visibility = 'extern',
                     api = 0, in_pxd = 0, is_cdef = 0,
-                    allow_pyobject = 0):
+                    allow_pyobject = 0, defining = 0):
         # Add an entry for an attribute.
         if not cname:
             cname = name
-        if type.is_cfunction:
-            type = PyrexTypes.CPtrType(type)
-        entry = self.declare(name, cname, type, pos, visibility)
+        entry = self.lookup_here(name)
+        if defining and entry is not None:
+            if not entry.type.same_as(type):
+                error(pos, "Function signature does not match previous declaration")
+        else:
+            entry = self.declare(name, cname, type, pos, visibility)
         entry.is_variable = 1
-        self.var_entries.append(entry)
+        if type.is_cfunction and self.type:
+            entry.func_cname = "%s::%s" % (self.type.declaration_code(""), cname)
+        if name != "this" and (defining or name != "<init>"):
+            self.var_entries.append(entry)
         if type.is_pyobject and not allow_pyobject:
             error(pos,
                 "C++ class member cannot be a Python object")
@@ -1891,10 +2036,12 @@ class CppClassScope(Scope):
         # Look for default constructors in all base classes.
         if self.default_constructor is None:
             entry = self.lookup(self.name)
-            if len(entry.type.base_classes) == 0:
+            if not entry.type.base_classes:
                 self.default_constructor = True
                 return
             for base_class in entry.type.base_classes:
+                if base_class is PyrexTypes.error_type:
+                    continue
                 temp_entry = base_class.scope.lookup_here("<init>")
                 found = False
                 if temp_entry is None:
@@ -1903,7 +2050,7 @@ class CppClassScope(Scope):
                     type = alternative.type
                     if type.is_ptr:
                         type = type.base_type
-                    if len(type.args) == 0:
+                    if not type.args:
                         found = True
                         break
                 if not found:
@@ -1917,16 +2064,23 @@ class CppClassScope(Scope):
     def declare_cfunction(self, name, type, pos,
                           cname = None, visibility = 'extern', api = 0, in_pxd = 0,
                           defining = 0, modifiers = (), utility_code = None):
-        if name == self.name.split('::')[-1] and cname is None:
+        if name in (self.name.split('::')[-1], '__init__') and cname is None:
             self.check_base_default_constructor(pos)
+            cname = self.type.cname
             name = '<init>'
-            type.return_type = self.lookup(self.name).type
+            type.return_type = PyrexTypes.InvisibleVoidType()
+        elif name == '__dealloc__' and cname is None:
+            cname = "~%s" % self.type.cname
+            name = '<del>'
+            type.return_type = PyrexTypes.InvisibleVoidType()
         prev_entry = self.lookup_here(name)
         entry = self.declare_var(name, type, pos,
+                                 defining=defining,
                                  cname=cname, visibility=visibility)
-        if prev_entry:
+        if prev_entry and not defining:
             entry.overloaded_alternatives = prev_entry.all_alternatives()
         entry.utility_code = utility_code
+        type.entry = entry
         return entry
 
     def declare_inherited_cpp_attributes(self, base_scope):
@@ -1958,9 +2112,10 @@ class CppClassScope(Scope):
         for entry in self.entries.values():
             if entry.is_type:
                 scope.declare_type(entry.name,
-                                    entry.type.specialize(values),
-                                    entry.pos,
-                                    entry.cname)
+                                   entry.type.specialize(values),
+                                   entry.pos,
+                                   entry.cname,
+                                   template=1)
             else:
 #                scope.declare_var(entry.name,
 #                                    entry.type.specialize(values),
@@ -1996,53 +2151,3 @@ class PropertyScope(Scope):
             error(pos, "Only __get__, __set__ and __del__ methods allowed "
                 "in a property declaration")
             return None
-
-
-# Should this go elsewhere (and then get imported)?
-#------------------------------------------------------------------------------------
-
-classmethod_utility_code = Code.UtilityCode(
-proto = """
-#include "descrobject.h"
-static PyObject* __Pyx_Method_ClassMethod(PyObject *method); /*proto*/
-""",
-impl = """
-static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
-    /* It appears that PyMethodDescr_Type is not anywhere exposed in the Python/C API */
-    static PyTypeObject *methoddescr_type = NULL;
-    if (methoddescr_type == NULL) {
-       PyObject *meth = __Pyx_GetAttrString((PyObject*)&PyList_Type, "append");
-       if (!meth) return NULL;
-       methoddescr_type = Py_TYPE(meth);
-       Py_DECREF(meth);
-    }
-    if (PyObject_TypeCheck(method, methoddescr_type)) { /* cdef classes */
-        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
-        #if PY_VERSION_HEX < 0x03020000
-        PyTypeObject *d_type = descr->d_type;
-        #else
-        PyTypeObject *d_type = descr->d_common.d_type;
-        #endif
-        return PyDescr_NewClassMethod(d_type, descr->d_method);
-    }
-    else if (PyMethod_Check(method)) { /* python classes */
-        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
-    }
-    else if (PyCFunction_Check(method)) {
-        return PyClassMethod_New(method);
-    }
-#ifdef __pyx_binding_PyCFunctionType_USED
-    else if (PyObject_TypeCheck(method, __pyx_binding_PyCFunctionType)) { /* binded CFunction */
-        return PyClassMethod_New(method);
-    }
-#endif
-    PyErr_Format(PyExc_TypeError,
-                 "Class-level classmethod() can only be called on "
-                 "a method_descriptor or instance method.");
-    return NULL;
-}
-""")
-
-#------------------------------------------------------------------------------------
-
-ERR_BUF_LOCALONLY = 'Buffer types only allowed as function local variables'
