@@ -3,54 +3,17 @@
 #
 #   Tree visitor and transform framework
 #
-import cython
 import inspect
+import TypeSlots
+import Builtin
 import Nodes
 import ExprNodes
-import Naming
 import Errors
 import DebugFlags
 
-class BasicVisitor(object):
-    """A generic visitor base class which can be used for visiting any kind of object."""
-    # Note: If needed, this can be replaced with a more efficient metaclass
-    # approach, resolving the jump table at module load time rather than per visitor
-    # instance.
-    def __init__(self):
-        self.dispatch_table = {}
+import cython
 
-    def visit(self, obj):
-        try:
-            handler_method = self.dispatch_table[type(obj)]
-        except KeyError:
-            handler_method = self.find_handler(obj)
-            self.dispatch_table[type(obj)] = handler_method
-        return handler_method(obj)
-
-    def find_handler(self, obj):
-        cls = type(obj)
-        #print "Cache miss for class %s in visitor %s" % (
-        #    cls.__name__, type(self).__name__)
-        # Must resolve, try entire hierarchy
-        pattern = "visit_%s"
-        mro = inspect.getmro(cls)
-        handler_method = None
-        for mro_cls in mro:
-            if hasattr(self, pattern % mro_cls.__name__):
-                handler_method = getattr(self, pattern % mro_cls.__name__)
-                break
-        if handler_method is None:
-            print type(self), cls
-            if hasattr(self, 'access_path') and self.access_path:
-                print self.access_path
-                if self.access_path:
-                    print self.access_path[-1][0].pos
-                    print self.access_path[-1][0].__dict__
-            raise RuntimeError("Visitor does not accept object: %s" % obj)
-        #print "Caching " + cls.__name__
-        return handler_method
-
-class TreeVisitor(BasicVisitor):
+class TreeVisitor(object):
     """
     Base class for writing visitors for a Cython tree, contains utilities for
     recursing such trees using visitors. Each node is
@@ -58,16 +21,16 @@ class TreeVisitor(BasicVisitor):
     containing child nodes or lists of child nodes. Lists are not considered
     part of the tree structure (i.e. contained nodes are considered direct
     children of the parent node).
-    
+
     visit_children visits each of the children of a given node (see the visit_children
     documentation). When recursing the tree using visit_children, an attribute
     access_path is maintained which gives information about the current location
     in the tree as a stack of tuples: (parent_node, attrname, index), representing
     the node, attribute and optional list index that was taken in each step in the path to
     the current node.
-    
+
     Example:
-    
+
     >>> class SampleNode(object):
     ...     child_attrs = ["head", "body"]
     ...     def __init__(self, value, head=None, body=None):
@@ -93,16 +56,17 @@ class TreeVisitor(BasicVisitor):
     out 3
     out 0
     """
-    
     def __init__(self):
         super(TreeVisitor, self).__init__()
+        self.dispatch_table = {}
         self.access_path = []
 
     def dump_node(self, node, indent=0):
         ignored = list(node.child_attrs) + [u'child_attrs', u'pos',
-                                            u'gil_message', u'subexprs']
+                                            u'gil_message', u'cpp_message',
+                                            u'subexprs']
         values = []
-        pos = node.pos
+        pos = getattr(node, 'pos', None)
         if pos:
             source = pos[0]
             if source:
@@ -143,7 +107,7 @@ class TreeVisitor(BasicVisitor):
                 code = frame.f_code
                 method_name = code.co_name
                 pos = (os.path.basename(code.co_filename),
-                       code.co_firstlineno)
+                       frame.f_lineno)
                 nodes.append((node, method_name, pos))
                 last_traceback = stacktrace
             stacktrace = stacktrace.tb_next
@@ -169,15 +133,53 @@ class TreeVisitor(BasicVisitor):
             trace.append(u"File '%s', line %d, in %s: %s" % (
                 pos[0], pos[1], method_name, self.dump_node(node)))
         raise Errors.CompilerCrash(
-            last_node.pos, self.__class__.__name__,
+            getattr(last_node, 'pos', None), self.__class__.__name__,
             u'\n'.join(trace), e, stacktrace)
 
-    def visitchild(self, child, parent, attrname, idx):
+    @cython.final
+    def find_handler(self, obj):
+        # to resolve, try entire hierarchy
+        cls = type(obj)
+        pattern = "visit_%s"
+        mro = inspect.getmro(cls)
+        handler_method = None
+        for mro_cls in mro:
+            handler_method = getattr(self, pattern % mro_cls.__name__, None)
+            if handler_method is not None:
+                return handler_method
+        print type(self), cls
+        if self.access_path:
+            print self.access_path
+            print self.access_path[-1][0].pos
+            print self.access_path[-1][0].__dict__
+        raise RuntimeError("Visitor %r does not accept object: %s" % (self, obj))
+
+    def visit(self, obj):
+        return self._visit(obj)
+
+    @cython.final
+    def _visit(self, obj):
+        try:
+            handler_method = self.dispatch_table[type(obj)]
+        except KeyError:
+            handler_method = self.find_handler(obj)
+            self.dispatch_table[type(obj)] = handler_method
+        return handler_method(obj)
+
+    @cython.final
+    def _visitchild(self, child, parent, attrname, idx):
         self.access_path.append((parent, attrname, idx))
         try:
-            result = self.visit(child)
+            try:
+                handler_method = self.dispatch_table[type(child)]
+            except KeyError:
+                handler_method = self.find_handler(child)
+                self.dispatch_table[type(child)] = handler_method
+            result = handler_method(child)
         except Errors.CompileError:
             raise
+        except Errors.AbortError:
+            raise
         except Exception, e:
             if DebugFlags.debug_no_exception_intercept:
                 raise
@@ -188,11 +190,12 @@ class TreeVisitor(BasicVisitor):
     def visitchildren(self, parent, attrs=None):
         return self._visitchildren(parent, attrs)
 
+    @cython.final
     def _visitchildren(self, parent, attrs):
         """
         Visits the children of the given parent. If parent is None, returns
         immediately (returning None).
-        
+
         The return value is a dictionary giving the results for each
         child (mapping the attribute name to either the return value
         or a list of return values (in the case of multiple children
@@ -205,9 +208,9 @@ class TreeVisitor(BasicVisitor):
             child = getattr(parent, attr)
             if child is not None:
                 if type(child) is list:
-                    childretval = [self.visitchild(x, parent, attr, idx) for idx, x in enumerate(child)]
+                    childretval = [self._visitchild(x, parent, attr, idx) for idx, x in enumerate(child)]
                 else:
-                    childretval = self.visitchild(child, parent, attr, None)
+                    childretval = self._visitchild(child, parent, attr, None)
                     assert not isinstance(childretval, list), 'Cannot insert list here: %s in %r' % (attr, parent)
                 result[attr] = childretval
         return result
@@ -217,14 +220,14 @@ class VisitorTransform(TreeVisitor):
     """
     A tree transform is a base class for visitors that wants to do stream
     processing of the structure (rather than attributes etc.) of a tree.
-    
+
     It implements __call__ to simply visit the argument node.
-    
+
     It requires the visitor methods to return the nodes which should take
     the place of the visited node in the result tree (which can be the same
     or one or more replacement). Specifically, if the return value from
     a visitor method is:
-    
+
     - [] or None; the visited node will be removed (set to None if an attribute and
     removed if in a list)
     - A single node; the visited node will be replaced by the returned node.
@@ -248,18 +251,18 @@ class VisitorTransform(TreeVisitor):
                         else:
                             newlist.append(x)
                 setattr(parent, attr, newlist)
-        return result        
+        return result
 
     def recurse_to_children(self, node):
         self.visitchildren(node)
         return node
-    
+
     def __call__(self, root):
-        return self.visit(root)
+        return self._visit(root)
 
 class CythonTransform(VisitorTransform):
     """
-    Certain common conventions and utilitues for Cython transforms.
+    Certain common conventions and utilities for Cython transforms.
 
      - Sets up the context of the pipeline in self.context
      - Tracks directives in effect in self.current_directives
@@ -287,9 +290,9 @@ class CythonTransform(VisitorTransform):
 
 class ScopeTrackingTransform(CythonTransform):
     # Keeps track of type of scopes
-    scope_type = None # can be either of 'module', 'function', 'cclass', 'pyclass'
-    scope_node = None
-    
+    #scope_type: can be either of 'module', 'function', 'cclass', 'pyclass', 'struct'
+    #scope_node: the node that owns the current scope
+
     def visit_ModuleNode(self, node):
         self.scope_type = 'module'
         self.scope_node = node
@@ -303,7 +306,7 @@ class ScopeTrackingTransform(CythonTransform):
         self.visitchildren(node)
         self.scope_type, self.scope_node = prev
         return node
-    
+
     def visit_CClassDefNode(self, node):
         return self.visit_scope(node, 'cclass')
 
@@ -319,18 +322,155 @@ class ScopeTrackingTransform(CythonTransform):
 
 class EnvTransform(CythonTransform):
     """
-    This transformation keeps a stack of the environments. 
+    This transformation keeps a stack of the environments.
     """
     def __call__(self, root):
-        self.env_stack = [root.scope]
-        return super(EnvTransform, self).__call__(root)        
-    
+        self.env_stack = [(root, root.scope)]
+        return super(EnvTransform, self).__call__(root)
+
+    def current_env(self):
+        return self.env_stack[-1][1]
+
+    def current_scope_node(self):
+        return self.env_stack[-1][0]
+
     def visit_FuncDefNode(self, node):
-        self.env_stack.append(node.local_scope)
+        self.env_stack.append((node, node.local_scope))
         self.visitchildren(node)
         self.env_stack.pop()
         return node
 
+    def visit_GeneratorBodyDefNode(self, node):
+        self.visitchildren(node)
+        return node
+
+    def visit_ClassDefNode(self, node):
+        self.env_stack.append((node, node.scope))
+        self.visitchildren(node)
+        self.env_stack.pop()
+        return node
+
+    def visit_CStructOrUnionDefNode(self, node):
+        self.env_stack.append((node, node.scope))
+        self.visitchildren(node)
+        self.env_stack.pop()
+        return node
+
+    def visit_ScopedExprNode(self, node):
+        if node.expr_scope:
+            self.env_stack.append((node, node.expr_scope))
+            self.visitchildren(node)
+            self.env_stack.pop()
+        else:
+            self.visitchildren(node)
+        return node
+
+
+class MethodDispatcherTransform(EnvTransform):
+    """
+    Base class for transformations that want to intercept on specific
+    builtin functions or methods of builtin types.  Must run after
+    declaration analysis when entries were assigned.
+
+    Naming pattern for handler methods is as follows:
+
+    * builtin functions: _handle_(general|simple|any)_function_NAME
+
+    * builtin methods: _handle_(general|simple|any)_method_TYPENAME_METHODNAME
+    """
+    # only visit call nodes
+    def visit_GeneralCallNode(self, node):
+        self.visitchildren(node)
+        function = node.function
+        if not function.type.is_pyobject:
+            return node
+        arg_tuple = node.positional_args
+        if not isinstance(arg_tuple, ExprNodes.TupleNode):
+            return node
+        keyword_args = node.keyword_args
+        if keyword_args and not isinstance(keyword_args, ExprNodes.DictNode):
+            # can't handle **kwargs
+            return node
+        args = arg_tuple.args
+        return self._dispatch_to_handler(
+            node, function, args, keyword_args)
+
+    def visit_SimpleCallNode(self, node):
+        self.visitchildren(node)
+        function = node.function
+        if function.type.is_pyobject:
+            arg_tuple = node.arg_tuple
+            if not isinstance(arg_tuple, ExprNodes.TupleNode):
+                return node
+            args = arg_tuple.args
+        else:
+            args = node.args
+        return self._dispatch_to_handler(
+            node, function, args)
+
+    ### dispatch to specific handlers
+
+    def _find_handler(self, match_name, has_kwargs):
+        call_type = has_kwargs and 'general' or 'simple'
+        handler = getattr(self, '_handle_%s_%s' % (call_type, match_name), None)
+        if handler is None:
+            handler = getattr(self, '_handle_any_%s' % match_name, None)
+        return handler
+
+    def _dispatch_to_handler(self, node, function, arg_list, kwargs=None):
+        if function.is_name:
+            # we only consider functions that are either builtin
+            # Python functions or builtins that were already replaced
+            # into a C function call (defined in the builtin scope)
+            if not function.entry:
+                return node
+            is_builtin = function.entry.is_builtin or\
+                         function.entry is self.current_env().builtin_scope().lookup_here(function.name)
+            if not is_builtin:
+                return node
+            function_handler = self._find_handler(
+                "function_%s" % function.name, kwargs)
+            if function_handler is None:
+                return node
+            if kwargs:
+                return function_handler(node, arg_list, kwargs)
+            else:
+                return function_handler(node, arg_list)
+        elif function.is_attribute and function.type.is_pyobject:
+            attr_name = function.attribute
+            self_arg = function.obj
+            obj_type = self_arg.type
+            is_unbound_method = False
+            if obj_type.is_builtin_type:
+                if obj_type is Builtin.type_type and arg_list and\
+                   arg_list[0].type.is_pyobject:
+                    # calling an unbound method like 'list.append(L,x)'
+                    # (ignoring 'type.mro()' here ...)
+                    type_name = function.obj.name
+                    self_arg = None
+                    is_unbound_method = True
+                else:
+                    type_name = obj_type.name
+            else:
+                type_name = "object" # safety measure
+            method_handler = self._find_handler(
+                "method_%s_%s" % (type_name, attr_name), kwargs)
+            if method_handler is None:
+                if attr_name in TypeSlots.method_name_to_slot\
+                or attr_name == '__new__':
+                    method_handler = self._find_handler(
+                        "slot%s" % attr_name, kwargs)
+                if method_handler is None:
+                    return node
+            if self_arg is not None:
+                arg_list = [self_arg] + list(arg_list)
+            if kwargs:
+                return method_handler(node, arg_list, kwargs, is_unbound_method)
+            else:
+                return method_handler(node, arg_list, is_unbound_method)
+        else:
+            return node
+
 
 class RecursiveNodeReplacer(VisitorTransform):
     """
@@ -348,15 +488,12 @@ class RecursiveNodeReplacer(VisitorTransform):
         else:
             return node
 
-
+def recursively_replace_node(tree, old_node, new_node):
+    replace_in = RecursiveNodeReplacer(old_node, new_node)
+    replace_in(tree)
 
 
 # Utils
-def ensure_statlist(node):
-    if not isinstance(node, Nodes.StatListNode):
-        node = Nodes.StatListNode(pos=node.pos, stats=[node])
-    return node
-
 def replace_node(ptr, value):
     """Replaces a node. ptr is of the form used on the access path stack
     (parent, attrname, listidx|None)
@@ -382,7 +519,7 @@ class PrintTree(TreeVisitor):
 
     def __call__(self, tree, phase=None):
         print("Parse tree dump at phase '%s'" % phase)
-        self.visit(tree)
+        self._visit(tree)
         return tree
 
     # Don't do anything about process_list, the defaults gives
